@@ -14,19 +14,15 @@
 
 #include <ctype.h>
 #include <limits.h>
-#include <fcntl.h>
-#include <unistd.h>
 
-#define SYS_DIR_ROOT_NAMES "/sys/kernel/capability/names"
-#define SYS_DIR_ROOT_CODES "/sys/kernel/capability/codes"
+#ifdef INCLUDE_GPERF_OUTPUT
+/* we need to include it after #define _GNU_SOURCE is set */
+#include INCLUDE_GPERF_OUTPUT
+#endif
 
 /* Maximum output text length (16 per cap) */
 #define CAP_TEXT_SIZE    (16*__CAP_MAXBITS)
 
-#define LIBCAP_EFF   01
-#define LIBCAP_INH   02
-#define LIBCAP_PER   04
-
 /*
  * Parse a textual representation of capabilities, returning an internal
  * representation.
@@ -84,9 +80,7 @@ static int lookupname(char const **strp)
 	*strp = str.constp;
 	return n;
     } else {
-	int fd;
-	char *tmp;
-	int c, n;
+	int c;
 	unsigned len;
 
 	for (len=0; (c = str.constp[len]); ++len) {
@@ -105,6 +99,7 @@ static int lookupname(char const **strp)
 	}
 #else /* ie., ndef GPERF_DOWNCASE */
 	char const *s;
+	unsigned n;
 
 	for (n = __CAP_BITS; n--; )
 	    if (_cap_names[n] && (s = namcmp(str.constp, _cap_names[n]))) {
@@ -113,25 +108,6 @@ static int lookupname(char const **strp)
 	    }
 #endif /* def GPERF_DOWNCASE */
 
-	asprintf(&tmp, SYS_DIR_ROOT_NAMES "/%s", str.constp);
-	str.constp += len;
-	tmp[sizeof(SYS_DIR_ROOT_NAMES) + len] = '\0';
-
-	fd = open(tmp, O_RDONLY);
-	free(tmp);
-
-	if (fd >= 0) {
-	    char ref[10];
-
-	    len = read(fd, ref, 9);
-	    close(fd);
-	    if (len > 0) {
-		ref[len] = '\0';
-		n = strtoul(ref, NULL, 10);
-		*strp = str.constp;
-		return n;
-	    }
-	}
 	return -1;   	/* No definition available */
     }
 }
@@ -316,24 +292,8 @@ char *cap_to_name(cap_value_t cap)
 # error Recompile with correctly sized numeric array
 #endif
 	char *tmp, *result;
-	int fd, len;
-
-	asprintf(&tmp, SYS_DIR_ROOT_CODES "/%u", cap);
-	fd = open(tmp, O_RDONLY);
-
-	if (fd >= 0) {
-	    len = read(fd, tmp, sizeof(SYS_DIR_ROOT_CODES));
-	    close(fd);
-	    if (len > 0) {
-		tmp[len-1] = '\0';
-	    } else {
-		goto default_to_numeric;
-	    }
-	} else {
-	default_to_numeric:
-	    sprintf(tmp, "%u", cap);
-	}
 
+	asprintf(&tmp, "%u", cap);
 	result = _libcap_strdup(tmp);
 	free(tmp);
 
@@ -406,54 +366,69 @@ char *cap_to_text(cap_t caps, ssize_t *length_p)
 
     memset(histo, 0, sizeof(histo));
 
-    for (n = cap_maxbits; n--; )
+    /* default prevailing state to the upper - unnamed bits */
+    for (n = cap_maxbits-1; n > __CAP_BITS; n--)
 	histo[getstateflags(caps, n)]++;
 
+    /* find which combination of capability sets shares the most bits
+       we bias to preferring non-set (m=0) with the >= 0 test. Failing
+       to do this causes strange things to happen with older systems
+       that don't know about bits 32+. */
     for (m=t=7; t--; )
-	if (histo[t] > histo[m])
+	if (histo[t] >= histo[m])
 	    m = t;
 
+    /* capture remaining bits - selecting m from only the unnamed bits,
+       we maximize the likelihood that we won't see numeric capability
+       values in the text output. */
+    while (n--)
+	histo[getstateflags(caps, n)]++;
+
     /* blank is not a valid capability set */
     p = sprintf(buf, "=%s%s%s",
 		(m & LIBCAP_EFF) ? "e" : "",
 		(m & LIBCAP_INH) ? "i" : "",
 		(m & LIBCAP_PER) ? "p" : "" ) + buf;
 
-    for (t = 8; t--; )
-	if (t != m && histo[t]) {
-	    *p++ = ' ';
-	    for (n = 0; n < cap_maxbits; n++)
-		if (getstateflags(caps, n) == t) {
-		    char *this_cap_name;
+    for (t = 8; t--; ) {
+	if (t == m || !histo[t]) {
+	    continue;
+	}
+	*p++ = ' ';
+	for (n = 0; n < cap_maxbits; n++) {
+	    if (getstateflags(caps, n) == t) {
+	        char *this_cap_name;
 
-		    this_cap_name = cap_to_name(n);
-		    if ((strlen(this_cap_name) + (p - buf)) > CAP_TEXT_SIZE) {
-			cap_free(this_cap_name);
-			errno = ERANGE;
-			return NULL;
-		    }
-		    p += sprintf(p, "%s,", this_cap_name);
+	        this_cap_name = cap_to_name(n);
+	        if ((strlen(this_cap_name) + (p - buf)) > CAP_TEXT_SIZE) {
 		    cap_free(this_cap_name);
-		}
-	    p--;
-	    n = t & ~m;
-	    if (n)
-		p += sprintf(p, "+%s%s%s",
-			     (n & LIBCAP_EFF) ? "e" : "",
-			     (n & LIBCAP_INH) ? "i" : "",
-			     (n & LIBCAP_PER) ? "p" : "");
-	    n = ~t & m;
-	    if (n)
-		p += sprintf(p, "-%s%s%s",
-			     (n & LIBCAP_EFF) ? "e" : "",
-			     (n & LIBCAP_INH) ? "i" : "",
-			     (n & LIBCAP_PER) ? "p" : "");
-	    if (p - buf > CAP_TEXT_SIZE) {
-		errno = ERANGE;
-		return NULL;
+		    errno = ERANGE;
+		    return NULL;
+	        }
+	        p += sprintf(p, "%s,", this_cap_name);
+	        cap_free(this_cap_name);
 	    }
 	}
-
+	p--;
+	n = t & ~m;
+	if (n) {
+	    p += sprintf(p, "+%s%s%s",
+			 (n & LIBCAP_EFF) ? "e" : "",
+			 (n & LIBCAP_INH) ? "i" : "",
+			 (n & LIBCAP_PER) ? "p" : "");
+	}
+	n = ~t & m;
+	if (n) {
+	    p += sprintf(p, "-%s%s%s",
+			 (n & LIBCAP_EFF) ? "e" : "",
+			 (n & LIBCAP_INH) ? "i" : "",
+			 (n & LIBCAP_PER) ? "p" : "");
+	}
+	if (p - buf > CAP_TEXT_SIZE) {
+	    errno = ERANGE;
+	    return NULL;
+	}
+    }
     _cap_debug("%s", buf);
     if (length_p) {
 	*length_p = p - buf;
